# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Unit test helpers for OpenHTF tests and phases.

This module provides some utility for unit testing OpenHTF test phases and
whole tests.  Primarily, otherwise difficult to mock mechanisms are mocked
for you, and there are a handful of convenience assertions that may be used
to write more readable (and less fragile against output format change) tests.

The primary class in this module is the TestCase class, which is a subclass
of unittest.TestCase that provides some extra utility.  Use it the same way
you would use unittest.TestCase.  See below for examples.

OpenHTF plugs are somewhat difficult to mock (because references are grabbed
at import time, you end up having to poke at framework internals to do this),
so there's a utility decorator for doing just this, @patch_plugs.  See below
for examples of how to use it.  Note that plugs are initialized once and torn
down once for a single method decorated with @patch_plugs (regardless of how
many phases or Test objects are yielded).  If you need new plug instances,
separate your test into separate decorated test* methods in your test case
(this is good practice anyway).

Lastly, while not implemented here, it's common to need to temporarily alter
configuration values for individual tests.  This can be accomplished with the
@conf.save_and_restore decorator (see docs in conf.py, examples below).

A few isolated examples, also see test/util/test_test.py for some usage:

  from openhtf.util import conf
  from openhtf.util import test

  import mytest  # Contains phases under test.

  class PhasesTest(test.TestCase):

    # Decorate with conf.save_and_restore to temporarily set conf values.
    # NOTE: This must come before yields_phases.
    @conf.save_and_restore(phase_variance='test_phase_variance')
    # Decorate the test* method with this to be able to yield a phase to run it.
    @test.yields_phases
    def test_first_phase(self):
      phase_record = yield mytest.first_phase
      # Check a measurement value.
      self.assertMeasured(phase_record, 'my_measurement', 'value')
      # Check that the measurement outcome was PASS.
      self.assertMeasurementPass(phase_record, 'my_measurement')

    @test.patch_plugs(mock_my_plug='my_plug.MyPlug')
    def test_second_phase(self, mock_my_plug):  # arg must match keyword above.
      # mock_my_plug is a MagicMock, and will be passed to yielded test phases
      # in place of an instance of my_plug.MyPlug.  You can access it here to
      # configure return values (and later to assert calls to plug methods).
      mock_my_plug.measure_voltage.return_value = 5.0

      # Trigger a phase (or openhtf.Test instance) to run by yielding it.  The
      # resulting PhaseRecord is yielded back (or TestRecord if you yielded an
      # openhtf.Test instance instead of a phase).
      phase_record = yield mytest.second_phase  # uses my_plug.MyPlug

      # Apply assertions to the output, probably using utility methods on self,
      # see below for an exhaustive list of such utility assertions.
      self.assertPhaseContinue(phase_record)

      # You can apply any assertions on the mocked plug here.
      mock_my_plug.measure_voltage.assert_called_once_with()

      # You can yield multiple phases/Test instances, but it's generally
      # cleaner and more readable to limit to a single yield per test case.

    @test.patch_plugs(mock_my_plug='my_plug.MyPlug')
    def test_multiple(self, mock_my_plug):
      # You can also yield an entire openhtf.Test() object.  If you do, you get
      # a TestRecord yielded back instead of a PhaseRecord.
      test_rec = yield openhtf.Test(mytest.first_phase, mytest.second_phase)

      # Some utility assertions are provided for operating on test records (see
      # below for a full list).
      self.assertTestPass(test_rec)

List of assertions that can be used with PhaseRecord results:

  assertPhaseContinue(phase_record)
  assertPhaseRepeat(phase_record)
  assertPhaseStop(phase_record)
  assertPhaseError(phase_record, exc_type=None)

List of assertions that can be used with TestRecord results:

  assertTestPass(test_rec)
  assertTestFail(test_rec)
  assertTestError(test_rec, exc_type=None)
  assertTestOutcomeCode(test_rec, code)

List of assertions that can be used with either PhaseRecords or TestRecords:

  assertMeasured(phase_or_test_rec, measurement, value=mock.ANY)
  assertNotMeasured(phase_or_test_rec, measurement)
  assertMeasurementPass(phase_or_test_rec, measurement)
  assertMeasurementFail(phase_or_test_rec, measurement)
"""

import collections
import functools
import inspect
import logging
import sys
import types
import unittest

import mock

import openhtf
from openhtf import plugs
from openhtf import util
from openhtf.core import measurements
from openhtf.core import phase_executor
from openhtf.core import test_descriptor
from openhtf.core import test_record
from openhtf.core import test_state
from openhtf.plugs import device_wrapping
from openhtf.util import logs, functions
import six

logs.CLI_LOGGING_VERBOSITY = 2


class InvalidTestError(Exception):
  """Raised when there's something invalid about a test."""


class PhaseOrTestIterator(collections.abc.Iterator):

  def __init__(self, iterator, mock_plugs):
    """Create an iterator for iterating over Tests or phases to run.

    Args:
      iterator: Child iterator to use for obtaining Tests or test phases, must
          be a generator.
      mock_plugs: Dict mapping plug types to mock objects to use instead of
          actually instantiating that type.

    Raises:
      InvalidTestError: when iterator is not a generator.
    """
    if not isinstance(iterator, types.GeneratorType):
      raise InvalidTestError(
          'Methods decorated with patch_plugs or yields_phases must yield '
          'test phases or openhtf.Test objects.', iterator)

    # Since we want to run single phases, we instantiate our own PlugManager.
    # Don't do this sort of thing outside OpenHTF unless you really know what
    # you're doing (http://imgur.com/iwBCmQe).
    self.plug_manager = plugs.PlugManager()
    self.iterator = iterator
    self.mock_plugs = mock_plugs
    self.last_result = None

  def _initialize_plugs(self, plug_types):
    # Make sure we initialize any plugs, this will ignore any that have already
    # been initialized.
    plug_types = list(plug_types)
    self.plug_manager.initialize_plugs(plug_cls for plug_cls in plug_types if
                                       plug_cls not in self.mock_plugs)
    for plug_type, plug_value in six.iteritems(self.mock_plugs):
      self.plug_manager.update_plug(plug_type, plug_value)

  def _handle_phase(self, phase_desc):
    """Handle execution of a single test phase."""
    logs.configure_logging()
    self._initialize_plugs(phase_plug.cls for phase_plug in phase_desc.plugs)

    # Cobble together a fake TestState to pass to the test phase.
    test_options = test_descriptor.TestOptions()
    with mock.patch(
        'openhtf.plugs.PlugManager', new=lambda _, __: self.plug_manager):
      test_state_ = test_state.TestState(
          openhtf.TestDescriptor((phase_desc,), phase_desc.code_info, {}),
          'Unittest:StubTest:UID',
          test_options
      )
      test_state_.mark_test_started()

    # Actually execute the phase, saving the result in our return value.
    executor = phase_executor.PhaseExecutor(test_state_)
    # Log an exception stack when a Phase errors out.
    with mock.patch.object(
        phase_executor.PhaseExecutorThread, '_log_exception',
        side_effect=logging.exception):
      # Use _execute_phase_once because we want to expose all possible outcomes.
      executor._execute_phase_once(phase_desc, is_last_repeat=False)
    return test_state_.test_record.phases[-1]

  def _handle_test(self, test):
    self._initialize_plugs(test.descriptor.plug_types)
    # Make sure we inject our mock plug instances.
    for plug_type, plug_value in six.iteritems(self.mock_plugs):
      self.plug_manager.update_plug(plug_type, plug_value)

    # We'll need a place to stash the resulting TestRecord.
    record_saver = util.NonLocalResult()
    test.add_output_callbacks(
        lambda record: setattr(record_saver, 'result', record))

    # Mock the PlugManager to use ours instead, and execute the test.
    with mock.patch(
        'openhtf.plugs.PlugManager', new=lambda _, __: self.plug_manager):
      test.execute(test_start=lambda: 'TestDutId')

    return record_saver.result

  def __next__(self):
    phase_or_test = self.iterator.send(self.last_result)
    if isinstance(phase_or_test, openhtf.Test):
      self.last_result = self._handle_test(phase_or_test)
    elif not isinstance(phase_or_test, collections.abc.Callable):
      raise InvalidTestError(
          'methods decorated with patch_plugs must yield Test instances or '
          'individual test phases', phase_or_test)
    else:
      self.last_result = self._handle_phase(
          openhtf.PhaseDescriptor.wrap_or_copy(phase_or_test))
    return phase_or_test, self.last_result

  def next(self):
    phase_or_test = self.iterator.send(self.last_result)
    if isinstance(phase_or_test, openhtf.Test):
      self.last_result = self._handle_test(phase_or_test)
    elif not isinstance(phase_or_test, collections.Callable):
      raise InvalidTestError(
          'methods decorated with patch_plugs must yield Test instances or '
          'individual test phases', phase_or_test)
    else:
      self.last_result = self._handle_phase(
          openhtf.PhaseDescriptor.wrap_or_copy(phase_or_test))
    return phase_or_test, self.last_result


def yields_phases(func):
  """Alias for patch_plugs with no plugs patched."""
  return patch_plugs()(func)


def patch_plugs(**mock_plugs):
  """Decorator for mocking plugs for a test phase.

  Usage:

    @plugs(my_plug=my_plug_module.MyPlug)
    def my_phase_that_uses_my_plug(test, my_plug):
      test.logger.info('Something: %s', my_plug.do_something(10))

    @test.patch_plugs(my_plug_mock='my_plug_module.MyPlug')
    def test_my_phase(self, my_plug_mock):
      # Set up return value for the do_something method on our plug.
      my_plug_mock.do_something.return_value = 'mocked_value'

      # Yield the phase you wish to test. Typically it wouldn't be in the same
      # module like this, but this works for example purposes.
      yield my_phase_that_uses_my_plug

      # Do some assertions to make sure our plug was used correctly.
      my_plug_mock.do_something.assert_called_with(10)

    # Passing in the plug class itself also works and can be beneficial
    # when the module path is long.
    @test.patch_plugs(my_plug_mock=my_plug_module.MyPlug)
    def test_my_phase_again(self, my_plug_mock):
      pass

  Args:
    **mock_plugs: kwargs mapping argument name to be passed to the test case to
        a string describing the plug type to mock.  The corresponding mock will
        be passed to the decorated test case as a keyword argument.

  Returns:
    Function decorator that mocks plugs.
  """
  def test_wrapper(test_func):
    plug_argspec = functions.getargspec(test_func)
    num_defaults = len(plug_argspec.defaults or ())
    plug_args = set(plug_argspec.args[1:-num_defaults or None])

    # Some sanity checks to make sure the mock arg names match.
    for arg in plug_args:
      if arg not in mock_plugs:
        raise InvalidTestError(
            'Test method %s expected arg %s, but it was not provided in '
            'patch_plugs kwargs: ' % (test_func.__name__, arg), mock_plugs)
    for mock_name in mock_plugs:
      if mock_name not in plug_args:
        raise InvalidTestError(
            'patch_plugs got kwarg %s, but test method %s does not expect '
            'it.' % (mock_name, test_func.__name__), plug_args)

    # Make MagicMock instances for the plugs.
    plug_kwargs = {}  # kwargs to pass to test func.
    plug_typemap = {}  # typemap for PlugManager, maps type to instance.
    for plug_arg_name, plug_fullname in six.iteritems(mock_plugs):
      if isinstance(plug_fullname, six.string_types):
        try:
          plug_module, plug_typename = plug_fullname.rsplit('.', 1)
          plug_type = getattr(sys.modules[plug_module], plug_typename)
        except Exception:
          logging.error("Invalid plug type specification %s='%s'",
                        plug_arg_name, plug_fullname)
          raise
      elif issubclass(plug_fullname, plugs.BasePlug):
        plug_type = plug_fullname
      else:
        raise ValueError('Invalid plug type specification %s="%s"' % (
            plug_arg_name, plug_fullname))
      if issubclass(plug_type, device_wrapping.DeviceWrappingPlug):
        # We can't strictly spec because calls to attributes are proxied to the
        # underlying device.
        plug_mock = mock.MagicMock()
      else:
        plug_mock = mock.create_autospec(plug_type, spec_set=True,
                                         instance=True)
      plug_typemap[plug_type] = plug_mock
      plug_kwargs[plug_arg_name] = plug_mock

    # functools.wraps is more than just aesthetic here, we need the original
    # name to match so we don't mess with unittest's TestLoader mechanism.
    @functools.wraps(test_func)
    def wrapped_test(self):
      for phase_or_test, result in PhaseOrTestIterator(
          test_func(self, **plug_kwargs), plug_typemap):
        logging.info('Ran %s, result: %s', phase_or_test, result)
    return wrapped_test
  return test_wrapper


class TestCase(unittest.TestCase):

  def __init__(self, methodName=None):
    super(TestCase, self).__init__(methodName=methodName)
    test_method = getattr(self, methodName)
    if inspect.isgeneratorfunction(test_method):
      raise ValueError(
          '%s yields without @openhtf.util.test.yields_phases' % methodName)

  def _AssertPhaseOrTestRecord(func):  # pylint: disable=no-self-argument,invalid-name
    """Decorator for automatically invoking self.assertTestPhases when needed.

    This allows assertions to apply to a single phase or "any phase in the test"
    without having to handle the type check themselves.  Note that the record,
    either PhaseRecord or TestRecord, must be the first argument to the
    wrapped assertion method.

    In the case of a TestRecord, the assertion will pass if *any* PhaseRecord in
    the TestRecord passes, otherwise the *last* exception raised will be
    re-raised.

    Returns:
      Function decorator.
    """
    @functools.wraps(func)
    def assertion_wrapper(self, phase_or_test_record, *args):
      if isinstance(phase_or_test_record, test_record.TestRecord):
        exc_info = None
        for phase_record in phase_or_test_record.phases:
          try:
            func(self, phase_record, *args)
            break
          except Exception:  # pylint: disable=broad-except
            exc_info = sys.exc_info()
        else:
          if exc_info:
            raise exc_info[0](exc_info[1]).raise_with_traceback(exc_info[2])
      elif isinstance(phase_or_test_record, test_record.PhaseRecord):
        func(self, phase_or_test_record, *args)
      else:
        raise InvalidTestError('Expected either a PhaseRecord or TestRecord')
    return assertion_wrapper

  ##### TestRecord Assertions #####

  def assertTestPass(self, test_rec):
    self.assertEqual(test_record.Outcome.PASS, test_rec.outcome)

  def assertTestFail(self, test_rec):
    self.assertEqual(test_record.Outcome.FAIL, test_rec.outcome)

  def assertTestAborted(self, test_rec):
    self.assertEqual(test_record.Outcome.ABORTED, test_rec.outcome)

  def assertTestError(self, test_rec, exc_type=None):
    self.assertEqual(test_record.Outcome.ERROR, test_rec.outcome)
    if exc_type:
      self.assertPhaseError(test_rec.phases[-1], exc_type)

  def assertTestOutcomeCode(self, test_rec, code):
    """Assert that the given code is in some OutcomeDetails in the record."""
    self.assertTrue(
        any(details.code == code for details in test_rec.outcome_details),
        'No OutcomeDetails had code %s' % code)

  ##### PhaseRecord Assertions #####

  def assertPhaseContinue(self, phase_record):
    self.assertIs(
        openhtf.PhaseResult.CONTINUE, phase_record.result.phase_result)

  def assertPhaseFailAndContinue(self, phase_record):
    self.assertIs(
        openhtf.PhaseResult.FAIL_AND_CONTINUE, phase_record.result.phase_result)

  def assertPhaseRepeat(self, phase_record):
    self.assertIs(openhtf.PhaseResult.REPEAT, phase_record.result.phase_result)

  def assertPhaseSkip(self, phase_record):
    self.assertIs(openhtf.PhaseResult.SKIP, phase_record.result.phase_result)

  def assertPhaseStop(self, phase_record):
    self.assertIs(openhtf.PhaseResult.STOP, phase_record.result.phase_result)

  def assertPhaseError(self, phase_record, exc_type=None):
    self.assertTrue(phase_record.result.raised_exception,
                    'Phase did not raise an exception')
    if exc_type:
      self.assertIsInstance(phase_record.result.phase_result.exc_val, exc_type,
                            'Raised exception %r is not a subclass of %r' %
                            (phase_record.result.phase_result, exc_type))

  def assertPhaseOutcomePass(self, phase_record):
    self.assertIs(test_record.PhaseOutcome.PASS, phase_record.outcome)

  def assertPhaseOutcomeFail(self, phase_record):
    self.assertIs(test_record.PhaseOutcome.FAIL, phase_record.outcome)

  def assertPhaseOutcomeSkip(self, phase_record):
    self.assertIs(test_record.PhaseOutcome.SKIP, phase_record.outcome)

  def assertPhaseOutcomeError(self, phase_record):
    self.assertIs(test_record.PhaseOutcome.ERROR, phase_record.outcome)

  ##### Measurement Assertions #####

  def assertNotMeasured(self, phase_or_test_record, measurement):
    def _check_phase(phase_record, strict=False):
      if strict:
        self.assertIn(measurement, phase_record.measurements)
      if measurement in phase_record.measurements:
        self.assertFalse(
            phase_record.measurements[measurement].measured_value.is_value_set,
            'Measurement %s unexpectedly set' % measurement)
        self.assertIs(measurements.Outcome.UNSET,
                      phase_record.measurements[measurement].outcome)

    if isinstance(phase_or_test_record, test_record.PhaseRecord):
      _check_phase(phase_or_test_record, True)
    else:
      # Check *all* phases (not *any* like _AssertPhaseOrTestRecord).
      for phase_record in phase_or_test_record.phases:
        _check_phase(phase_record)

  @_AssertPhaseOrTestRecord
  def assertMeasured(self, phase_record, measurement, value=mock.ANY):
    self.assertTrue(
        phase_record.measurements[measurement].measured_value.is_value_set,
        'Measurement %s not set' % measurement)
    if value is not mock.ANY:
      self.assertEqual(
          value, phase_record.measurements[measurement].measured_value.value,
          'Measurement %s has wrong value: expected %s, got %s' %
          (measurement, value,
           phase_record.measurements[measurement].measured_value.value))

  @_AssertPhaseOrTestRecord
  def assertMeasurementPass(self, phase_record, measurement):
    self.assertMeasured(phase_record, measurement)
    self.assertIs(measurements.Outcome.PASS,
                  phase_record.measurements[measurement].outcome)

  @_AssertPhaseOrTestRecord
  def assertMeasurementFail(self, phase_record, measurement):
    self.assertMeasured(phase_record, measurement)
    self.assertIs(measurements.Outcome.FAIL,
                  phase_record.measurements[measurement].outcome)



import os
import sys
from glob import glob
from setuptools import setup, find_packages
from setuptools.command.install import install as install_orig
from setuptools.command.develop import develop as develop_orig


HERE = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(HERE, 'src', 'spintop_openhtf', 'VERSION')) as version_file:
    VERSION = version_file.read().strip()

# if not os.path.exists(os.path.join('openhtf', 'openhtf')):
#     raise Exception('Please checkout the openhtf submodule before install.')

compiled_proto_files = glob('src/openhtf/output/proto/*_pb2.py')
if not compiled_proto_files:
    raise Exception('Protobuf files in openhtf where not compiled. Must be done before hand.')

# Find packages under openhtf(git) / openhtf
# Add the package itself

# openhtf_packages = ['openhtf'] + ['openhtf.' + pack for pack in find_packages(where='openhtf/openhtf')]
packages =  find_packages('src')

setup(
    name='spintop-openhtf',
    version=VERSION,
    description='A complete integration to spintop.io using OpenHTF. OpenHTF is currently vendored-in.',
    author='William Laroche',
    author_email='william.laroche@tackv.ca',
    maintainer='William Laroche',
    maintainer_email='william.laroche@tackv.ca',
    url='https://gitlab.com/tackv/spintop-openhtf',
    package_dir={
        '': 'src',
    },
    packages=packages,
    package_data={
        'openhtf': [
            'examples/example_attachment.txt',
            'examples/example_config.yaml'
            'output/proto/*.proto',
            'output/web_gui/dist/*',
            'output/web_gui/dist/css/*',
            'output/web_gui/dist/js/*',
            'output/web_gui/dist/img/*',
            'output/web_gui/*'
        ],
        'spintop_openhtf': [
            'VERSION',
            'callbacks/web_gui/dist/*',
            'callbacks/web_gui/dist/css/*',
            'callbacks/web_gui/dist/js/*',
            'callbacks/web_gui/dist/img/*',
            'callbacks/web_gui/*'
        ]
    },
    install_requires=[
        'appdirs>=1.0.0',
        'oauth2client>=4.1.0',
        'colorama>=0.3.9,<1.0',
        'contextlib2>=0.5.1,<1.0',
        'future>=0.16.0',
        'gspread>=3.1.0',
        'google-api-python-client>=1.7.10',
        'mutablerecords>=0.4.1,<2.0',
        'oauth2client>=4.1.3',
        'protobuf>=3.6.0,<4.0',
        'PyYAML>=5.0',
        'pytz',
        'pyOpenSSL>=17.1.0,<18.0',
        'SheetFu>=1.4.1',
        'jsonschema>=3.0.2',
        'markdown2>=2.3.8'
    ],
    extras_require={
        'server':[ # Allows the different GUI servers. Can still run without GUI.
            'tornado>=4.3,<5.0',
            'sockjs-tornado>=1.0.3,<2.0',
        ],
        'plugs.comport':[
            'pyserial>=3.3.0,<4.0',
        ],
        'plugs.ssh':[
            'paramiko',
        ],
        'plugs.visa':[
            'pyvisa',
        ]
    },
    setup_requires=[
        'wheel>=0.29.0,<1.0',
    ],
    tests_require=[
        'mock>=2.0.0',
        'pytest>=2.9.2',
        'pytest-cov>=2.2.1',
    ],
)


from datetime import datetime, timedelta
from openhtf.core.test_record import TestRecord

from spintop.models import SpintopTestRecord
from spintop.utils import utcnow_aware

from spintop_openhtf.transforms.openhtf_fmt import OpenHTFTestRecordTransformer
from spintop_openhtf.transforms.openhtf_gen import OpenHTFTestsGenerator


generator = OpenHTFTestsGenerator()
transformer = OpenHTFTestRecordTransformer()

generate_transform = generator + transformer

def test_generate():
    tests = list(generator.generate(count=2))

    assert len(tests) == 2
    assert isinstance(tests[0], TestRecord)

def test_transform():

    now = utcnow_aware()
    min_datetime = now - timedelta(minutes=1)
    max_datetime = now + timedelta(minutes=1)


    spintop_records = list(generate_transform.generate(count=2))

    assert len(spintop_records) == 2
    assert isinstance(spintop_records[0], SpintopTestRecord)
    
    for record in spintop_records:
        assert min_datetime < record.test_id.start_datetime < max_datetime




import random
import time

import openhtf as htf
from openhtf.output.callbacks import json_factory

from openhtf.util import validators

from spintop.generators import Generator

def normal(mu, sigma):
    return random.normalvariate(mu, sigma)

@htf.measures('first', 'second', 'third', 'fourth')
def lots_of_measurements(test, base_mu, base_sigma):
    for index, measurement in enumerate(('first', 'second', 'third', 'fourth')):
        test.measurements[measurement] = normal(base_mu - ((5*index)%13), base_sigma+(index))

@htf.PhaseOptions()
def sleep_sometime(test, sleep_length):
    if sleep_length:
        time.sleep(sleep_length)

@htf.PhaseOptions()
def random_failure(test, failure_odd):
    if random.uniform(0, 1.0) < failure_odd:
        return htf.PhaseResult.STOP
        
@htf.measures('something_else')
def singular_test(test):
    test.measurements.something_else = 6.2
    

base_mu_gen = lambda index: 5+(index*0.25)
base_sigma_gen = lambda index: 2
test_name_gen = lambda index: 'basicgen_test'
dut_id_gen = lambda index: 'dut%d' % ((index%25) if 17 < (index%25) < 20 else index % 17)
additionnal_phases_gen = lambda index: [singular_test] if (index%25) == 7 else []
sleep_gen = lambda index: 1 if (index%25) == 10 or (index%25) == 17 else None


class OpenHTFTestsGenerator(Generator):
    def __call__(self, count=25):
        # We instantiate our OpenHTF test with the phases we want to run as args.
        self.tests = []
        def callback(test_record):
            self.tests.append(test_record)
        
        for index in range(count):
            
            phases = [
                lots_of_measurements.with_args(base_mu=base_mu_gen(index), base_sigma=base_sigma_gen(index)),
                random_failure.with_args(failure_odd=0.1),
                sleep_sometime.with_args(sleep_length=sleep_gen(index))
            ] + additionnal_phases_gen(index)
            
            test = htf.Test(*phases, test_name=test_name_gen(index))

            test.add_output_callbacks(callback)

            dut_id = dut_id_gen(index)
            test.execute(test_start=lambda: dut_id)
        
        for test in self.tests:
            # Requires a generator
            yield test
    
    
  
    


""" Standardized additions to openhtf """
import enum

#
# STANDARD CONF
#

def init_conf():
    pass


#
# ENVIRONMENT
#

EnvironmentType = enum.Enum('EnvironmentType', [
    'DEVELOPMENT',
    'PRODUCTION'
])

def get_env():
    # TODO : something more advanced
    return EnvironmentType.DEVELOPMENT

def is_development_env():
    return get_env() == EnvironmentType.DEVELOPMENT

def is_production_env():
    return get_env() == EnvironmentType.PRODUCTION
    




from .base import UnboundPlug


import os
import time

def bytes_from_file(filename, chunksize=8192):
    with open(filename, "rb") as f:
        while True:
            chunk = f.read(chunksize)
            if chunk:
                for b in chunk:
                    yield b
            else:
                break

class VISAException(Exception):
    pass

class VISAInterface(UnboundPlug):
    def __init__(self, instrument=None, open_timeout=5, timeout=5):
        super().__init__()
        self.instr_name = instrument
        self.open_timeout = open_timeout*1000 # ms
        self.timeout = timeout*1000 # ms
        self.instr = None

    # Old opentest "Server" Side Interface

    def open(self, _instrument=None):
        self.close()
        if not _instrument:
            import pyvisa
            self.logger.info("Connecting to instrument [%s]", self.instr_name)
            try:
                resource_manager = pyvisa.ResourceManager()
            except OSError as e:
                raise VISAException("Please install NI-VISA driver on this computer in order to use the VISA Service.") from e
            _instrument = resource_manager.open_resource(self.instr_name, open_timeout=self.open_timeout)
            _instrument.timeout = self.timeout

        self.instr = _instrument

    def close(self):
        if self.instr is not None:
            self.instr.close()
            self.instr = None

    def visa_client_connected(function):
        def check_connected(self, *args, **kwargs):
            if self.instr is None:
                self.logger.info("Connection is no longer alive")
                self.connect()
            return function(self, *args, **kwargs)

        return check_connected

    @visa_client_connected
    def _execute_request(self, 
            multiple_commands=[], 
            command="", 
            read_binary=False,
            file_name="",
            doread=False,
            unstack_errors=False,
            timeout=None
        ):

        if timeout is None: timeout = self.timeout

        self.logger.debug("Timeout is %s seconds" % timeout)
        timeout = timeout*1000 # to ms
        self.instr.timeout = timeout
        output = ""
        if file_name != "" and command != "":
            if read_binary:
                self.logger.info('Copying remote file to %s' % file_name)
                values = self.instr.query_binary_values(command, datatype='c')
                values = bytearray(values)
                with open(file_name, 'wb') as f:
                    f.write(values)
            else:
                #Command + binary file
                binary = bytes_from_file(file_name)
                self.logger.info("Sending binary file: %s<%s>" % (command, file_name))
                self.instr.write_binary_values(command, list(binary), datatype='c')
        else:
            #Command only
            if command != "":
                multiple_commands = [command] + multiple_commands

            output = self._try_command_x_times(self._send_commands, 3, multiple_commands, doread)

        if unstack_errors:
            self.assert_no_errors()

        return output

    def _send_commands(self, multiple_commands, doread):
        output = ''
        for cmd in multiple_commands:
            self.logger.info("Sending command: %s", cmd)
            self._try_command_x_times(self.instr.write, 3, cmd)

        try:
            if doread:
                output = self.instr.read()
                self.logger.info("Read: %s", output.strip())
            else:
                #If not expected to read, block until all operations are completed
                self.logger.debug("Waiting for operations to finish...")
                output = self.instr.query("*OPC?")
        except Exception as e:
            self.logger.info(e)
            self.logger.warning("Intrument read failed")
            self.logger.debug("Unstacking all errors from instrument")
            raise VISAException(self._format_get_errors_read())

        return output.strip()

    def _try_command_x_times(self, fn, tries, *args, **kwargs):
        while True:
            try:
                return fn(*args, **kwargs)
            except Exception as e:
                tries = tries - 1
                if tries <= 0:
                    raise
                self.logger.warning('%s: Retrying' % str(e))

    def assert_no_errors(self):
        errors = self._get_errors()
        if len(errors) > 0:
            raise VISAException(self._format_get_errors_read(errors))

    def _format_get_errors_read(self, errors=None):
        if not errors:
            errors = self._get_errors()
        if len(errors) <= 0:
            string = "No errors but read failed: was there something to be read?"
        else:
            string = "ERRORS"
            for error in errors:
                string = string + ":" + error

        return string

    def _get_errors(self):
        ERROR_QUERY = "SYST:ERR?"
        errors = []
        error_string = self.instr.query(ERROR_QUERY)
        self.logger.debug("%s", error_string.strip())
        code, msg = error_string.split(',')
        code = code.strip()
        msg = msg.strip()
        if int(code) != 0:
            errors = errors + [msg]
            errors = errors + self._get_errors()

        return errors

    # Old opentest "Client" Side Interface

    def execute_command(self, command, doread=False, unstack_errors=False, timeout=None):
        return self._execute_request(command=command, doread=doread, unstack_errors=unstack_errors, timeout=timeout)

    def write_binary_file(self, command, file_name, **kwargs):
        output = self._execute_request(command=command, file_name=file_name, **kwargs)

    def read_binary_file(self, command, file_name, **kwargs):
        output = self._execute_request(command=command, file_name=file_name, read_binary=True, **kwargs)

    def queue_command(self, command):
        self.command_queue = self.command_queue + [command]

    def execute_queued_commands(self, **kwargs):
        try:
            output = self._execute_request(multiple_commands=self.command_queue, **kwargs)
        finally:
            self.command_queue = []
        return output

    def set_and_read_command(self, command, set_value):
        set_cmd = command + " " + str(set_value)
        read_cmd = command + "?"
        return self.execute_command(set_cmd + ";" + read_cmd, doread = True)

    def read(self):
        return self._execute_request(doread=True)

    def read_error(self):
        return self._execute_request(unstack_errors=True)

    def _parse_float_output(self, output):
        values_string = output.split(',')
        values = ()
        for value_str in values_string:
            values = values + (float(value_str),)
        return values

    def reset(self):
        self.execute_command('*RST', timeout=10.0)
        self.instrument_reset()

    def instrument_reset(self):
        pass



# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Fastboot device."""

import logging
import time

from openhtf.plugs.usb import fastboot_protocol
from openhtf.plugs.usb import usb_exceptions
from openhtf.util import timeouts

# From fastboot.c
VENDORS = {0x18D1, 0x0451, 0x0502, 0x0FCE, 0x05C6, 0x22B8, 0x0955,
           0x413C, 0x2314, 0x0BB4, 0x8087}
CLASS = 0xFF
SUBCLASS = 0x42
PROTOCOL = 0x03

_LOG = logging.getLogger(__name__)
class FastbootDevice(object):
  """Libusb fastboot wrapper with retries."""

  def __init__(self, _, num_retries=3):
    self._num_retries = num_retries
    self._protocol = None

  @property
  def usb_handle(self):
    """Return our USB handle."""
    return self._protocol.usb_handle

  def set_boot_config(self, name, value):
    """Set the boot configuration."""
    self.oem('bootconfig %s %s' % (name, value))

  def get_boot_config(self, name, info_cb=None):
    """Get bootconfig, either as full dict or specific value for key."""
    result = {}
    def default_info_cb(msg):
      """Default Info CB."""
      if not msg.message:
        return
      key, value = msg.message.split(':', 1)
      result[key.strip()] = value.strip()
    info_cb = info_cb or default_info_cb
    final_result = self.oem('bootconfig %s' % name, info_cb=info_cb)
    # Return INFO messages before the final OKAY message.
    if name in result:
      return result[name]
    return final_result

  def lock(self):
    """Lock the device."""
    self.oem('lock', timeout_ms=1000)

  def close(self):
    """Close the device."""
    if self._protocol:
      self.__getattr__('Close')()
      self._protocol = None

  def __getattr__(self, attr):  # pylint: disable=invalid-name
    """Fallthrough to underlying FastbootProtocol handler.

    Args:
      attr: Attribute to get.
    Returns:
      Either the attribute from the device or a retrying function-wrapper
      if attr is a method on the device.
    """
    if not self._protocol:
      raise usb_exceptions.HandleClosedError()

    val = getattr(self._protocol, attr)
    if callable(val):
      def _retry_wrapper(*args, **kwargs):
        """Wrap the retry function."""
        result = _retry_usb_function(self._num_retries, val, *args, **kwargs)
        _LOG.debug('LIBUSB FASTBOOT: %s(*%s, **%s) -> %s',
                   attr, args, kwargs, result)
        return result
      return _retry_wrapper
    return val

  @classmethod
  def connect(cls, usb_handle, **kwargs):
    """Connect to the device.

    Args:
      usb_handle: UsbHandle instance to use for communication to the device.
      **kwargs: Additional args to pass to the class constructor (currently
          only num_retries).

    Returns:
      An instance of this class if the device connected successfully.
    """
    return cls(fastboot_protocol.FastbootCommands(usb_handle), **kwargs)

def _retry_usb_function(count, func, *args, **kwargs):
  """Helper function to retry USB."""
  helper = timeouts.RetryHelper(count)
  while True:
    try:
      return func(*args, **kwargs)
    except usb_exceptions.CommonUsbError:
      if not helper.retry_if_possible():
        raise
      time.sleep(0.1)
    else:
      break



# Copyright 2015 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""User input module for OpenHTF.

Provides a plug which can be used to prompt the user for input. The prompt can
be displayed in the console, the OpenHTF web GUI, and custom frontends.
"""

from __future__ import print_function

import collections
import functools
import logging
import os
import platform
import select
import sys
import threading
import uuid
import time
import enum

from openhtf import PhaseOptions
from openhtf import plugs
from openhtf.util import console_output, conf
from six.moves import input

if platform.system() != 'Windows':
  import termios  # pylint: disable=g-import-not-at-top

_LOG = logging.getLogger(__name__)

PROMPT = '--> '

PromptType = enum.Enum('PromptType', [
  'OKAY',
  'FORM',
  'OKAY_CANCEL',
  'PASS_FAIL',
  'PASS_FAIL_RETRY'
])

conf.declare(
  'user_input_enable_console', 
  'If True, enables user input collection in console prompt.', 
  default_value=True
)

class InvalidOption(Exception):
  def __init__(self, option):
    self.option = option
    super(InvalidOption, self).__init__('Option %s is not valid' % self.option)

class SecondaryOptionOccured(Exception):
  pass

class CancelOption(SecondaryOptionOccured):
  pass

class FailOption(SecondaryOptionOccured):
  pass

class RetryOption(SecondaryOptionOccured):
  pass

class PromptOption(object):
  def __init__(self, key, error=None, name=None, error_level=False, warning_level=False, success_level=False):
    self.key = key
    self.error = error
    
    if (error and not warning_level) or error_level:
      self.level = 'error'
    elif warning_level:
      self.level = 'warning'
    elif success_level:
      self.level = 'success'
    else:
      self.level = 'info'
      
    if name is None:
      name = key
    self.name = name
    
  def as_dict(self):
    return {'key': self.key, 'name': self.name, 'level': self.level}
    
  def raise_if_error(self):
    if self.error:
      raise self.error
  
class PromptResponse(object):
  def __init__(self, response, prompt_type):
    self.content = str(response)
    self.option = None
    
    if isinstance(response, collections.Mapping):
      self.content = response.get('content', '')
      self.option = response.get('option', None)
    # elif prompt_type != PromptType.FORM:
    #   self.option = response
    
    if not self.option:
      self.option = None
    elif isinstance(self.option, PromptOption):
      self.option = self.option.key
    
    if self.option not in OPTIONS:
      raise InvalidOption(self.option)
    
    self.option = OPTIONS[self.option]
    
  def return_value(self):
    self.option.raise_if_error()
    return self.content
      

OKAY = PromptOption('OKAY')

OPTIONS = {
  None: OKAY,
  'OKAY': OKAY,
  'SUBMIT': PromptOption('SUBMIT', name='Submit'),
  'CANCEL': PromptOption('CANCEL', CancelOption, warning_level=True),
  'PASS': PromptOption('PASS', success_level=True),
  'FAIL': PromptOption('FAIL', FailOption),
  'RETRY': PromptOption('RETRY', RetryOption, warning_level=True)
}

PROMPT_TO_OPTIONS = {
  PromptType.OKAY: [OPTIONS['OKAY']],
  PromptType.FORM: [OPTIONS['SUBMIT']],
  PromptType.OKAY_CANCEL: [OPTIONS['OKAY'], OPTIONS['CANCEL']],
  PromptType.PASS_FAIL: [OPTIONS['PASS'], OPTIONS['FAIL']]
}

def create_text_input_form(input_message='Input'):
  return {
      'schema': {
          'title': "Generic Input",
          'type': "object",
          'required': ["_input"],
          'properties': {
              '_input': {
                  'type': "string", 
                  'title': input_message
              }
          }
      }
  }

class PromptInputError(Exception):
  """Raised in the event that a prompt returns without setting the response."""


class MultiplePromptsError(Exception):
  """Raised if a prompt is invoked while there is an existing prompt."""


class PromptUnansweredError(Exception):
  """Raised when a prompt times out or otherwise comes back unanswered."""


Prompt = collections.namedtuple('Prompt', 'id message prompt_type')

CONSOLE_PROMPT = None

class ConsolePrompt(threading.Thread):
  """Thread that displays a prompt to the console and waits for a response.

  This should not be used for processes that run in the background.
  """

  def __init__(self, message, callback, color=''):
    """Initializes a ConsolePrompt.

    Args:
      message: A string to be presented to the user.
      callback: A function to be called with the response string.
      color: An ANSI color code, or the empty string.
    """
    super(ConsolePrompt, self).__init__()
    self.daemon = True
    self._message = str(message)
    self._callback = callback
    self._color = color
    self._stop_event = threading.Event()
    self._answered = False

  def Stop(self):
    """Mark this ConsolePrompt as stopped."""
    self._stop_event.set()
    if not self._answered:
      _LOG.debug('Stopping ConsolePrompt--prompt was answered from elsewhere.')

  def run(self):
    """Main logic for this thread to execute."""
    if platform.system() == 'Windows':
      # Windows doesn't support file-like objects for select(), so fall back
      # to raw_input().
      response = input(''.join((self._message,
                                os.linesep,
                                PROMPT)))
      self._answered = True
      self._callback(response)
      return

    # First, display the prompt to the console.
    console_output.cli_print(self._message, color=self._color,
                             end=os.linesep, logger=None)
    console_output.cli_print(PROMPT, color=self._color, end='', logger=None)
    sys.stdout.flush()

    # Before reading, clear any lingering buffered terminal input.
    termios.tcflush(sys.stdin, termios.TCIFLUSH)

    line = ''
    while not self._stop_event.is_set():
      inputs, _, _ = select.select([sys.stdin], [], [], 0.001)
      if sys.stdin in inputs:
        new = os.read(sys.stdin.fileno(), 1024)
        if not new:
          # Hit EOF!
          # They hit ^D (to insert EOF). Tell them to hit ^C if they
          # want to actually quit.
          print('Hit ^C (Ctrl+c) to exit.')
          break
        line += new.decode('utf-8')
        if '\n' in line:
          response = line[:line.find('\n')]
          self._answered = True
          self._callback(response)
          return


def prompt_type_console_message(options=[]):
  if len(options) <= 1:
    return "Press enter to continue."
  else:
    return "Select one of the following option: " + " or ".join(opt.key for opt in options)
  
def prompt_options(prompt_type):
  if prompt_type in PROMPT_TO_OPTIONS:
    return PROMPT_TO_OPTIONS[prompt_type]
  else:
    raise RuntimeError('PromptType %s does not exists.' % prompt_type)
 
class UserInput(plugs.FrontendAwareBasePlug):
  """Get user input from inside test phases.

  Attributes:
    last_response: None, or a pair of (prompt_id, response) indicating the last
        user response that was received by the plug.
  """

  def __init__(self):
    super(UserInput, self).__init__()
    self.last_response = None
    self._prompt = None
    self._response = None
    self._cond = threading.Condition(threading.RLock())

  @property
  def _console_prompt(self):
    """ Console prompt is global because it uses the only available once stdin to capture input.
    If not global, a new test with create a new plug and any old console prompt will be left hanging.""" 
    return CONSOLE_PROMPT

  @_console_prompt.setter
  def _console_prompt(self, value):
    global CONSOLE_PROMPT
    CONSOLE_PROMPT = value

  def _asdict(self):
    """Return a dictionary representation of the current prompt."""
    with self._cond:
      if self._prompt is None:
        return
      return {'id': self._prompt.id,
              'message': self._prompt.message,
              'prompt_type': self._prompt.prompt_type,
              'options': [opt.as_dict() for opt in self._options]}

  def tearDown(self):
    self.remove_prompt()

  def remove_prompt(self):
    """Remove the prompt."""
    with self._cond:
      self._prompt = None
      if self._console_prompt:
        self._console_prompt.Stop()
      self.notify_update()

  def prompt_form(self, json_schema_form, prompt_type=PromptType.OKAY, timeout_s=None):
    message = json_schema_form
    return self.prompt(message, prompt_type=prompt_type, timeout_s=timeout_s)
    
  def prompt(self, message, prompt_type=PromptType.OKAY, text_input=False, cli_color='', timeout_s=None):
    """Display a prompt and wait for a response.

    Args:
      message: A string to be presented to the user.
      prompt_type: The type of prompt:
        OKAY: Simple info prompt with an okay button.
        OKAY_CANCEL: Okay or Cancel buttons.
        PASS_FAIL: PASS or FAIL buttons.
        PASS_FAIL_RETRY: PASS, FAIL, RETRY buttons. (not working)
        
      timeout_s: Seconds to wait before raising a PromptUnansweredError.
      cli_color: An ANSI color code, or the empty string.

    Returns:
      A string response, or the empty string if text_input was False.

    Raises:
      MultiplePromptsError: There was already an existing prompt.
      PromptUnansweredError: Timed out waiting for the user to respond.
    """
    if text_input:
      ret_val = self.prompt_form(create_text_input_form(message))
      try:
        return ret_val['_input']
      except (TypeError, KeyError):
        return ret_val
    
    self.start_prompt(message, prompt_type, cli_color)
    try:
      return self.wait_for_prompt(timeout_s)
    except InvalidOption as e:
      _LOG.info('Option %s is invalid. Please retry.' % e.option)
      return self.prompt(message, prompt_type, cli_color, timeout_s)

  def start_prompt(self, message, prompt_type=PromptType.OKAY, cli_color=''):
    """Display a prompt.

    Args:
      message: A string to be presented to the user.
      prompt_type: The type of prompt:
        OKAY: Simple info prompt with an okay button.
        OKAY_CANCEL: Okay or Cancel buttons.
        PASS_FAIL: PASS or FAIL buttons.
        PASS_FAIL_RETRY: PASS, FAIL, RETRY buttons. (not working)
      cli_color: An ANSI color code, or the empty string.

    Raises:
      MultiplePromptsError: There was already an existing prompt.

    Returns:
      A string uniquely identifying the prompt.
    """
    with self._cond:
      if self._prompt:
        raise MultiplePromptsError
      
      prompt_id = uuid.uuid4().hex
      _LOG.debug('Displaying prompt (%s): "%s"', prompt_id, message)

      self._response = None
      self._prompt = Prompt(
          id=prompt_id, message=message, prompt_type=prompt_type)
      self._options = prompt_options(prompt_type)
      self._prompt_type = prompt_type
      if conf['user_input_enable_console'] and sys.stdin.isatty():
        console_message = str(message) + ', ' + prompt_type_console_message(options=self._options)
        
        if self._console_prompt and self._console_prompt.is_alive():
          # Console prompt still active; replacing callback with new prompt.
          # Prolly on Windows since the prompt stop function does not interrupt input()
          self._console_prompt._callback = functools.partial(self.respond, prompt_id)
        else:
          self._console_prompt = ConsolePrompt(
              console_message, functools.partial(self.respond, prompt_id), cli_color)
          self._console_prompt.start()

      self.notify_update()
      return prompt_id

  def wait_for_prompt(self, timeout_s=None):
    """Wait for the user to respond to the current prompt.

    Args:
      timeout_s: Seconds to wait before raising a PromptUnansweredError.

    Returns:
      A string response, or the empty string if text_input was False.

    Raises:
      PromptUnansweredError: Timed out waiting for the user to respond.
    """
    start_time = time.time()

    with self._cond:
      # Waiting in short increments allows sys interrupts to be captured.
      # This allows Ctrl-C to be used while waiting for a prompt.
      while self._prompt and not self._cond.wait(0.5):
        # If no timeout, we just wait continuously
        if timeout_s is not None and time.time() - start_time > timeout_s:
          break

    if self._response is None:
      raise PromptUnansweredError
    
    response = PromptResponse(self._response, self._prompt_type)
    return response.return_value()

  def respond(self, prompt_id, response):
    """Respond to the prompt with the given ID.

    If there is no active prompt or the given ID doesn't match the active
    prompt, do nothing.

    Args:
      prompt_id: A string uniquely identifying the prompt.
      response: A string response to the given prompt.

    Returns:
      True if the prompt with the given ID was active, otherwise False.
    """
    _LOG.debug('Responding to prompt (%s): "%s"', prompt_id, response)
    with self._cond:
      if not (self._prompt and self._prompt.id == prompt_id):
        return False
      
      self._response = response
      self.last_response = (prompt_id, response)
      self.remove_prompt()
      self._cond.notifyAll()
    return True


def prompt_for_test_start(
    message='Enter a DUT ID in order to start the test.', timeout_s=60*60*24,
    validator=lambda sn: sn, cli_color=''):
  """Returns an OpenHTF phase for use as a prompt-based start trigger.

  Args:
    message: The message to display to the user.
    timeout_s: Seconds to wait before raising a PromptUnansweredError.
    validator: Function used to validate or modify the serial number.
    cli_color: An ANSI color code, or the empty string.
  """

  @PhaseOptions(timeout_s=timeout_s)
  @plugs.plug(prompts=UserInput)
  def trigger_phase(test, prompts):
    """Test start trigger that prompts the user for a DUT ID."""
    dut_id = prompts.prompt(
        message, text_input=True, timeout_s=timeout_s, cli_color=cli_color)
    test.test_record.dut_id = validator(dut_id)

  return trigger_phase



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: openhtf/output/proto/assembly_event.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2
from google.protobuf import struct_pb2 as google_dot_protobuf_dot_struct__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='openhtf/output/proto/assembly_event.proto',
  package='openhtf',
  syntax='proto3',
  serialized_pb=_b('\n)openhtf/output/proto/assembly_event.proto\x12\x07openhtf\x1a\x1fgoogle/protobuf/timestamp.proto\x1a\x1cgoogle/protobuf/struct.proto\"\xb9\x04\n\rAssemblyEvent\x12\x30\n\x06target\x18\x01 \x01(\x0b\x32 .openhtf.AssemblyEvent.Component\x12/\n\x05\x63hild\x18\x03 \x01(\x0b\x32 .openhtf.AssemblyEvent.Component\x12,\n\x02op\x18\x04 \x01(\x0e\x32 .openhtf.AssemblyEvent.Operation\x12(\n\x04time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x1a\x96\x02\n\tComponent\x12\x13\n\x0bpart_number\x18\x01 \x01(\t\x12\x10\n\x06serial\x18\x02 \x01(\tH\x00\x12\x35\n\x03lot\x18\x03 \x01(\x0b\x32&.openhtf.AssemblyEvent.Component.ByLotH\x00\x12\x19\n\x11\x65lectronic_serial\x18\x04 \x01(\t\x12\x15\n\rinstance_name\x18\x05 \x01(\t\x12+\n\nextra_data\x18\x06 \x01(\x0b\x32\x17.google.protobuf.Struct\x12\x16\n\x0e\x64\x65viation_code\x18\x07 \x03(\t\x1a.\n\x05\x42yLot\x12\x12\n\nlot_number\x18\x01 \x01(\t\x12\x11\n\tlot_index\x18\x02 \x01(\tB\x04\n\x02id\"T\n\tOperation\x12\n\n\x06\x41TTACH\x10\x00\x12\n\n\x06\x44\x45TACH\x10\x01\x12\n\n\x06\x43REATE\x10\x02\x12\n\n\x06UPDATE\x10\x03\x12\x17\n\x13\x44\x45TACH_ALL_CHILDREN\x10\x04\"\x89\x02\n\x13StaticComponentInfo\x12\x1f\n\x17\x65xtraction_timestamp_ms\x18\x01 \x01(\x03\x12!\n\x19modification_timestamp_ms\x18\x02 \x01(\x03\x12\x0e\n\x06serial\x18\x03 \x01(\t\x12\x13\n\x0bpart_number\x18\x04 \x01(\t\x12\x15\n\rinstance_name\x18\x05 \x01(\t\x12\x12\n\ndeviations\x18\x06 \x03(\t\x12\x14\n\x0cupdate_count\x18\x07 \x01(\x03\x12\x33\n\tcomponent\x18\x08 \x01(\x0b\x32 .openhtf.AssemblyEvent.Component\x12\x13\n\x0b\x64\x65scription\x18\t \x01(\tb\x06proto3')
  ,
  dependencies=[google_dot_protobuf_dot_timestamp__pb2.DESCRIPTOR,google_dot_protobuf_dot_struct__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)



_ASSEMBLYEVENT_OPERATION = _descriptor.EnumDescriptor(
  name='Operation',
  full_name='openhtf.AssemblyEvent.Operation',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='ATTACH', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DETACH', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='CREATE', index=2, number=2,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='UPDATE', index=3, number=3,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='DETACH_ALL_CHILDREN', index=4, number=4,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=603,
  serialized_end=687,
)
_sym_db.RegisterEnumDescriptor(_ASSEMBLYEVENT_OPERATION)


_ASSEMBLYEVENT_COMPONENT_BYLOT = _descriptor.Descriptor(
  name='ByLot',
  full_name='openhtf.AssemblyEvent.Component.ByLot',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='lot_number', full_name='openhtf.AssemblyEvent.Component.ByLot.lot_number', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lot_index', full_name='openhtf.AssemblyEvent.Component.ByLot.lot_index', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=549,
  serialized_end=595,
)

_ASSEMBLYEVENT_COMPONENT = _descriptor.Descriptor(
  name='Component',
  full_name='openhtf.AssemblyEvent.Component',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='part_number', full_name='openhtf.AssemblyEvent.Component.part_number', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='serial', full_name='openhtf.AssemblyEvent.Component.serial', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='lot', full_name='openhtf.AssemblyEvent.Component.lot', index=2,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='electronic_serial', full_name='openhtf.AssemblyEvent.Component.electronic_serial', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='openhtf.AssemblyEvent.Component.instance_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='extra_data', full_name='openhtf.AssemblyEvent.Component.extra_data', index=5,
      number=6, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='deviation_code', full_name='openhtf.AssemblyEvent.Component.deviation_code', index=6,
      number=7, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ASSEMBLYEVENT_COMPONENT_BYLOT, ],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
    _descriptor.OneofDescriptor(
      name='id', full_name='openhtf.AssemblyEvent.Component.id',
      index=0, containing_type=None, fields=[]),
  ],
  serialized_start=323,
  serialized_end=601,
)

_ASSEMBLYEVENT = _descriptor.Descriptor(
  name='AssemblyEvent',
  full_name='openhtf.AssemblyEvent',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='target', full_name='openhtf.AssemblyEvent.target', index=0,
      number=1, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='child', full_name='openhtf.AssemblyEvent.child', index=1,
      number=3, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='op', full_name='openhtf.AssemblyEvent.op', index=2,
      number=4, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='time', full_name='openhtf.AssemblyEvent.time', index=3,
      number=5, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[_ASSEMBLYEVENT_COMPONENT, ],
  enum_types=[
    _ASSEMBLYEVENT_OPERATION,
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=118,
  serialized_end=687,
)


_STATICCOMPONENTINFO = _descriptor.Descriptor(
  name='StaticComponentInfo',
  full_name='openhtf.StaticComponentInfo',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='extraction_timestamp_ms', full_name='openhtf.StaticComponentInfo.extraction_timestamp_ms', index=0,
      number=1, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='modification_timestamp_ms', full_name='openhtf.StaticComponentInfo.modification_timestamp_ms', index=1,
      number=2, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='serial', full_name='openhtf.StaticComponentInfo.serial', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='part_number', full_name='openhtf.StaticComponentInfo.part_number', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='instance_name', full_name='openhtf.StaticComponentInfo.instance_name', index=4,
      number=5, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='deviations', full_name='openhtf.StaticComponentInfo.deviations', index=5,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='update_count', full_name='openhtf.StaticComponentInfo.update_count', index=6,
      number=7, type=3, cpp_type=2, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='component', full_name='openhtf.StaticComponentInfo.component', index=7,
      number=8, type=11, cpp_type=10, label=1,
      has_default_value=False, default_value=None,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='openhtf.StaticComponentInfo.description', index=8,
      number=9, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=690,
  serialized_end=955,
)

_ASSEMBLYEVENT_COMPONENT_BYLOT.containing_type = _ASSEMBLYEVENT_COMPONENT
_ASSEMBLYEVENT_COMPONENT.fields_by_name['lot'].message_type = _ASSEMBLYEVENT_COMPONENT_BYLOT
_ASSEMBLYEVENT_COMPONENT.fields_by_name['extra_data'].message_type = google_dot_protobuf_dot_struct__pb2._STRUCT
_ASSEMBLYEVENT_COMPONENT.containing_type = _ASSEMBLYEVENT
_ASSEMBLYEVENT_COMPONENT.oneofs_by_name['id'].fields.append(
  _ASSEMBLYEVENT_COMPONENT.fields_by_name['serial'])
_ASSEMBLYEVENT_COMPONENT.fields_by_name['serial'].containing_oneof = _ASSEMBLYEVENT_COMPONENT.oneofs_by_name['id']
_ASSEMBLYEVENT_COMPONENT.oneofs_by_name['id'].fields.append(
  _ASSEMBLYEVENT_COMPONENT.fields_by_name['lot'])
_ASSEMBLYEVENT_COMPONENT.fields_by_name['lot'].containing_oneof = _ASSEMBLYEVENT_COMPONENT.oneofs_by_name['id']
_ASSEMBLYEVENT.fields_by_name['target'].message_type = _ASSEMBLYEVENT_COMPONENT
_ASSEMBLYEVENT.fields_by_name['child'].message_type = _ASSEMBLYEVENT_COMPONENT
_ASSEMBLYEVENT.fields_by_name['op'].enum_type = _ASSEMBLYEVENT_OPERATION
_ASSEMBLYEVENT.fields_by_name['time'].message_type = google_dot_protobuf_dot_timestamp__pb2._TIMESTAMP
_ASSEMBLYEVENT_OPERATION.containing_type = _ASSEMBLYEVENT
_STATICCOMPONENTINFO.fields_by_name['component'].message_type = _ASSEMBLYEVENT_COMPONENT
DESCRIPTOR.message_types_by_name['AssemblyEvent'] = _ASSEMBLYEVENT
DESCRIPTOR.message_types_by_name['StaticComponentInfo'] = _STATICCOMPONENTINFO

AssemblyEvent = _reflection.GeneratedProtocolMessageType('AssemblyEvent', (_message.Message,), dict(

  Component = _reflection.GeneratedProtocolMessageType('Component', (_message.Message,), dict(

    ByLot = _reflection.GeneratedProtocolMessageType('ByLot', (_message.Message,), dict(
      DESCRIPTOR = _ASSEMBLYEVENT_COMPONENT_BYLOT,
      __module__ = 'openhtf.output.proto.assembly_event_pb2'
      # @@protoc_insertion_point(class_scope:openhtf.AssemblyEvent.Component.ByLot)
      ))
    ,
    DESCRIPTOR = _ASSEMBLYEVENT_COMPONENT,
    __module__ = 'openhtf.output.proto.assembly_event_pb2'
    # @@protoc_insertion_point(class_scope:openhtf.AssemblyEvent.Component)
    ))
  ,
  DESCRIPTOR = _ASSEMBLYEVENT,
  __module__ = 'openhtf.output.proto.assembly_event_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.AssemblyEvent)
  ))
_sym_db.RegisterMessage(AssemblyEvent)
_sym_db.RegisterMessage(AssemblyEvent.Component)
_sym_db.RegisterMessage(AssemblyEvent.Component.ByLot)

StaticComponentInfo = _reflection.GeneratedProtocolMessageType('StaticComponentInfo', (_message.Message,), dict(
  DESCRIPTOR = _STATICCOMPONENTINFO,
  __module__ = 'openhtf.output.proto.assembly_event_pb2'
  # @@protoc_insertion_point(class_scope:openhtf.StaticComponentInfo)
  ))
_sym_db.RegisterMessage(StaticComponentInfo)


# @@protoc_insertion_point(module_scope)



import yaml

try:
    from jinja2 import Template 
except ImportError:
    pass

def _parse_net_string(net):
    refered_nets = net.split(',')
    return refered_nets

def load_yml_file(filename, *args, **kwargs):
    with open(filename) as open_file:
        return load_yml(open_file.read(), *args, **kwargs)

def load_yml(yml_string, use_jinja2=False, render_data={}):
    if use_jinja2:
        rendered = render_jinja2_template(yml_string, data=render_data)
    else:
        rendered = yml_string
    content = yaml.load(rendered)
    return content

def render_jinja2_template(content, data={}):
    template = Template(content)
    return template.render(**data)


# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Modified by Tack Verification, 2020

"""TestExecutor executes tests."""

import logging
import sys
import threading

from openhtf.core import phase_descriptor
from openhtf.core import phase_executor
from openhtf.core import phase_group
from openhtf.core import test_record
from openhtf.core import test_state
from openhtf.util import conf
from openhtf.util import threads


_LOG = logging.getLogger(__name__)


conf.declare('cancel_timeout_s', default_value=2,
             description='Timeout (in seconds) when the test has been cancelled'
             'to wait for the running phase to exit.')

conf.declare('stop_on_first_failure', default_value=False,
             description='Stop current test execution and return Outcome FAIL'
             'on first phase with failed measurement.')


class TestExecutionError(Exception):
  """Raised when there's an internal error during test execution."""


class TestStopError(Exception):
  """Test is being stopped."""


# pylint: disable=too-many-instance-attributes
class TestExecutor(threads.KillableThread):
  """Encompasses the execution of a single test."""
  daemon = True

  def __init__(self, test_descriptor, execution_uid, test_start, test_options):
    super(TestExecutor, self).__init__(name='TestExecutorThread')
    self.test_state = None

    self._test_descriptor = test_descriptor
    self._test_start = test_start
    self._test_options = test_options
    self._lock = threading.Lock()
    self._phase_exec = None
    self.uid = execution_uid
    self._last_outcome = None
    self._abort = threading.Event()
    self._full_abort = threading.Event()
    self._teardown_phases_lock = threading.Lock()

  def close(self):
    """Close and remove any global registrations.

    Always call this function when finished with this instance.

    This function is defined instead of a __del__ function because Python calls
    the __del__ function unreliably.
    """
    self.wait()
    self.test_state.close()

  def abort(self):
    """Abort this test."""
    if self._abort.is_set():
      _LOG.error('Abort already set; forcibly stopping the process.')
      self._full_abort.set()
      self._stop_phase_executor(force=True)
      return
    _LOG.error('Abort test executor.')
    # Deterministically mark the test as aborted.
    self._abort.set()
    self._stop_phase_executor()
    # No need to kill this thread because the abort state has been set, it will
    # end as soon as all queued teardown phases are run.

  def finalize(self):
    """Finalize test execution and output resulting record to callbacks.

    Should only be called once at the conclusion of a test run, and will raise
    an exception if end_time_millis is already set.

    Returns:
      Finalized TestState.  It should not be modified after this call.

    Raises:
      TestStopError: test
      TestAlreadyFinalized if end_time_millis already set.
    """
    if not self.test_state:
      raise TestStopError('Test Stopped.')
    if self.test_state.test_record.dut_id is None:
      _LOG.warning('DUT ID is still not set; using default.')
      self.test_state.test_record.dut_id = self._test_options.default_dut_id

    return self.test_state

  def wait(self):
    """Waits until death."""
    # Must use a timeout here in case this is called from the main thread.
    # Otherwise, the SIGINT abort logic in test_descriptor will not get called.

    # This used to call join with threading.TIMEOUT_MAX. However, on raspbian
    # this caused the join to return immediately. Changed to always use 1 year.
    
    while self.is_alive():
      threads.join_timeout_increments(self)
      
    # Sanity check to make sure we did not hit the one year timeout
    if self.is_alive():
        raise TestExecutionError('test_executor.wait() returned before thread died')

  def _thread_proc(self):
    """Handles one whole test from start to finish."""
    try:
      # Top level steps required to run a single iteration of the Test.
      self.test_state = test_state.TestState(
          self._test_descriptor,
          self.uid,
          self._test_options)
      phase_exec = phase_executor.PhaseExecutor(self.test_state)

      # Any access to self._exit_stacks must be done while holding this lock.
      with self._lock:
        self._phase_exec = phase_exec

      if self._test_start is not None and self._execute_test_start():
        # Exit early if test_start returned a terminal outcome of any kind.
        return
      self.test_state.mark_test_started()

      # Full plug initialization happens _after_ the start trigger, as close to
      # test execution as possible, for the best chance of test equipment being
      # in a known-good state at the start of test execution.
      if self._initialize_plugs():
        return

      # Everything is set, set status and begin test execution.
      self.test_state.set_status_running()
      self._execute_phase_group(self._test_descriptor.phase_group)
    except:
      _LOG.exception('Exception in TestExecutor.')
      raise
    finally:
      self._execute_test_teardown()

  def _initialize_plugs(self, plug_types=None):
    """Initialize plugs.

    Args:
      plug_types: optional list of plug classes to initialize.

    Returns:
      True if there was an error initializing the plugs.
    """
    try:
      self.test_state.plug_manager.initialize_plugs(plug_types=plug_types)
      return False
    except Exception:  # pylint: disable=broad-except
      # Record the equivalent failure outcome and exit early.
      self._last_outcome = phase_executor.PhaseExecutionOutcome(
          phase_executor.ExceptionInfo(*sys.exc_info()))
      return True

  def _execute_test_start(self):
    """Run the start trigger phase, and check that the DUT ID is set after.

    Initializes any plugs used in the trigger.
    Logs a warning if the start trigger failed to set the DUT ID.

    The test start is special because we wait to initialize all other plugs
    until this phase runs.

    Returns:
      True if there was a terminal error either setting up or running the test
      start phase.
    """
    # Have the phase executor run the start trigger phase. Do partial plug
    # initialization for just the plugs needed by the start trigger phase.
    if self._initialize_plugs(plug_types=[
        phase_plug.cls for phase_plug in self._test_start.plugs]):
      return True

    outcome = self._phase_exec.execute_phase(self._test_start)
    if outcome.is_terminal:
      self._last_outcome = outcome
      return True

    if self.test_state.test_record.dut_id is None:
      _LOG.warning('Start trigger did not set a DUT ID.')
    return False

  def _stop_phase_executor(self, force=False):
    with self._lock:
      phase_exec = self._phase_exec
      if not phase_exec:
        # The test executor has not started yet, so no stopping is required.
        return
    if not force and not self._teardown_phases_lock.acquire(False):
      # If locked, teardown phases are running, so do not cancel those.
      return
    try:
      phase_exec.stop(timeout_s=conf.cancel_timeout_s)
      # Resetting so phase_exec can run teardown phases.
      phase_exec.reset_stop()
    finally:
      if not force:
        self._teardown_phases_lock.release()

  # TODO(kschiller): Cleanup the naming here and possibly merge with finalize.
  def _execute_test_teardown(self):
    # Plug teardown does not affect the test outcome.
    self.test_state.plug_manager.tear_down_plugs()

    # Now finalize the test state.
    if self._abort.is_set():
      self.test_state.state_logger.debug('Finishing test with outcome ABORTED.')
      self.test_state.abort()
    elif self._last_outcome and self._last_outcome.is_terminal:
      self.test_state.finalize_from_phase_outcome(self._last_outcome)
    else:
      self.test_state.finalize_normally()

  def _handle_phase(self, phase):
    if isinstance(phase, phase_group.PhaseGroup):
      return self._execute_phase_group(phase)

    self.test_state.state_logger.debug('Handling phase %s', phase.name)
    outcome = self._phase_exec.execute_phase(phase)
    if (self.test_state.test_options.stop_on_first_failure or
        conf.stop_on_first_failure):
      # Stop Test on first measurement failure
      current_phase_result = self.test_state.test_record.phases[
          len(self.test_state.test_record.phases) - 1]
      if current_phase_result.outcome == test_record.PhaseOutcome.FAIL:
        outcome = phase_executor.PhaseExecutionOutcome(
            phase_descriptor.PhaseResult.STOP)
        self.test_state.state_logger.error(
            'Stopping test because stop_on_first_failure is True')

    if outcome.is_terminal and not self._last_outcome:
      self._last_outcome = outcome

    return outcome.is_terminal

  def _execute_abortable_phases(self, type_name, phases, group_name):
    """Execute phases, returning immediately if any error or abort is triggered.

    Args:
      type_name: str, type of phases running, usually 'Setup' or 'Main'.
      phases: iterable of phase_descriptor.Phase or phase_group.PhaseGroup
          instances, the phases to execute.
      group_name: str or None, name of the executing group.

    Returns:
      True if there is a terminal error or the test is aborted, False otherwise.
    """
    if group_name and phases:
      self.test_state.state_logger.debug(
          'Executing %s phases for %s', type_name, group_name)
    for phase in phases:
      if self._abort.is_set() or self._handle_phase(phase):
        return True
    return False

  def _execute_teardown_phases(self, teardown_phases, group_name):
    """Execute all the teardown phases, regardless of errors.

    Args:
      teardown_phases: iterable of phase_descriptor.Phase or
          phase_group.PhaseGroup instances, the phases to execute.
      group_name: str or None, name of the executing group.

    Returns:
      True if there is at least one terminal error, False otherwise.
    """
    if group_name and teardown_phases:
      self.test_state.state_logger.debug('Executing teardown phases for %s',
                                         group_name)
    ret = False
    with self._teardown_phases_lock:
      for teardown_phase in teardown_phases:
        if self._full_abort.is_set():
          ret = True
          break
        if self._handle_phase(teardown_phase):
          ret = True
    return ret

  def _execute_phase_group(self, group):
    """Executes the phases in a phase group.

    This will run the phases in the phase group, ensuring if the setup
    phases all run without error that the teardown phases will also run, no
    matter the errors during the main phases.

    This function is recursive.  Do not construct phase groups that contain
    themselves.

    Args:
      group: phase_group.PhaseGroup, the phase group to execute.

    Returns:
      True if the phases are terminal; otherwise returns False.
    """
    if group.name:
      self.test_state.state_logger.debug('Entering PhaseGroup %s', group.name)
    if self._execute_abortable_phases(
        'setup', group.setup, group.name):
      return True
    main_ret = self._execute_abortable_phases(
        'main', group.main, group.name)
    teardown_ret = self._execute_teardown_phases(
        group.teardown, group.name)
    return main_ret or teardown_ret



# Copyright 2016 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Example OpenHTF test logic.

Run with (your virtualenv must be activated first):

  python with_plugs.py

with_plugs() is most useful when you have a test that has to happen in the same
(or mostly similar) ways across multiple interfaces, such as testing
connectivity on a quad-NIC board. Instead of creating 4 phases with 4 plugs, you
can create 1 phase with 4 subclasses of the same plug and use with_plugs() to
end up with the 4 phases you want.
"""

from __future__ import print_function
import subprocess
import time

import openhtf as htf
from openhtf import plugs


class PingPlug(plugs.BasePlug):
  """This plug simply does a ping against the host attribute."""
  host = None

  def __init__(self):
    # This should only ever be constructed by a subclass, so this is only
    # checking that the subclass set host correctly.
    assert self.host is not None

  def _get_command(self, count):
    # Returns the commandline for pinging the host.
    return [
      'ping',
      '-c',
      str(count),
      self.host,
    ]

  def run(self, count):
    command = self._get_command(count)
    print("running: %s" % ' '.join(command))
    return subprocess.call(command)


# These subclasses specify the host to ping so they can be used directly by
# phases or through with_plugs().
class PingGoogle(PingPlug):
  host = 'google.com'


class PingDnsA(PingPlug):
  host = '8.8.8.8'


class PingDnsB(PingPlug):
  host = '8.8.4.4'


# Note: phase name and total_time measurement use {} formatting with args
# passed into the phase so each phase has a unique name.
@htf.PhaseOptions(name='Ping-{pinger.host}-{count}')
@htf.plug(pinger=PingPlug.placeholder)
@htf.measures(
    'total_time_{pinger.host}_{count}',
    htf.Measurement('retcode').equals('{expected_retcode}', type=int)
)
def test_ping(test, pinger, count, expected_retcode):
  """This tests that we can ping a host.

  The plug, pinger, is expected to be replaced at test creation time, so the
  placeholder property was used instead of the class directly.
  """
  del expected_retcode  # Not used in the phase, only used by a measurement
  start = time.time()
  retcode = pinger.run(count)
  elapsed = time.time() - start
  test.measurements['total_time_%s_%s' % (pinger.host, count)] = elapsed
  test.measurements.retcode = retcode


if __name__ == '__main__':
  # We instantiate our OpenHTF test with the phases we want to run as args.

  # We're going to use these these plugs to create all our phases using only 1
  # written phase.
  ping_plugs = [
    PingGoogle,
    PingDnsA,
    PingDnsB,
  ]
  
  phases = [
      test_ping.with_plugs(pinger=plug).with_args(count=2, expected_retcode=0)
      for plug in ping_plugs
  ]

  test = htf.Test(*phases)

  # Unlike hello_world.py, where we prompt for a DUT ID, here we'll just
  # use an arbitrary one.
  test.execute(test_start=lambda: 'MyDutId')



# Copyright 2018 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""Simple OpenHTF test which launches the web GUI client."""

import openhtf as htf
from openhtf.util import conf

from openhtf.output.servers import station_server
from openhtf.output.web_gui import web_launcher
from openhtf.plugs import user_input


@htf.measures(htf.Measurement('hello_world_measurement'))
def hello_world(test):
  test.logger.info('Hello World!')
  test.measurements.hello_world_measurement = 'Hello Again!'


if __name__ == '__main__':
  conf.load(station_server_port='4444')
  with station_server.StationServer() as server:
    web_launcher.launch('http://localhost:4444')
    for i in range(5):
      test = htf.Test(hello_world)
      test.add_output_callbacks(server.publish_final_state)
      test.execute(test_start=user_input.prompt_for_test_start())



import logging
import gspread
import datetime

from ..util.gdrive import FileNotFound, load_credentials_file, GoogleDrive, Worksheet, GoogleFolder
from openhtf.core.test_record import Outcome
from openhtf import conf


SCOPE = ['https://spreadsheets.google.com/feeds',
         'https://www.googleapis.com/auth/drive']
            
INDEX_FILE = 'index'
TESTS_TAB = 'Tests'
MEASUREMENTS_TAB = 'Measurements'
CONFIG_TAB = 'Config'
LOGS_TAB = 'Logs'

def load_credentials_file_with_scope(credentials_file):
    return load_credentials_file(credentials_file, SCOPE)

def create_google_drive_outputs(credentials, folder_name=None, share_with=[]):

    def on_output(test_record):
        if (test_record.outcome == Outcome.ABORTED
                and test_record.dut_id == conf.default_dut_id):
            # skip export
            pass
        else:
            gwrap = GoogleDrive(credentials)
            folder_id = gwrap.get_or_create_folder(folder_name, share_with=share_with)
            folder = GoogleFolder(gwrap, folder_id)

            outputs = [
                TestRecordToSheetOutput(folder),
                TestRecordToRowOutput(folder, INDEX_FILE)
            ]

            for output in outputs:
                output(test_record)

    return [on_output]

class TestRecordToSheetOutput(object):
    """ Creates a new google sheet in the constructor-argument folder.
    The file is named '{dut_id}-{timestamp}'.
    There are four tabs:
        - Tests
        - Measurements
        - Config
        - Logs
    """
    TEST_PHASE_COLUMNS = [
        ('name', lambda phase: phase.name),
        ('start_time', lambda phase: str(millis_to_datetime(phase.start_time_millis))),
        ('end_time', lambda phase: str(millis_to_datetime(phase.end_time_millis))),
        ('outcome', lambda phase: phase.outcome.name),
    ]

    def __init__(self, folder):
        self.folder = folder

    def __call__(self, test_record):
        filename = '{dut_id}-{timestamp}'.format(
            dut_id=test_record.dut_id, 
            timestamp=str(millis_to_datetime(test_record.start_time_millis))
        )
        
        spreadsheet = self.folder.get_or_create_spreadsheet(filename)
        file_id = spreadsheet.id
        file_url = 'https://docs.google.com/spreadsheets/d/%s/edit?usp=drivesdk' % file_id
        test_record.metadata['google_sheet_result_file_url'] = file_url
        self.write_test_phases(spreadsheet, test_record)
        self.write_test_measurements(spreadsheet, test_record)
        self.write_test_config(spreadsheet, test_record)
        self.write_logs(spreadsheet, test_record)

    def write_test_phases(self, spreadsheet, test_record):
        """ Writes the 'Tests' tab. """
        tab = Worksheet(spreadsheet, TESTS_TAB, headers=self.get_phase_header(), first=True)
        tab.append_rows([self.get_phase_values(phase) for phase in test_record.phases])

    def get_phase_header(self):
        return [col[0] for col in self.TEST_PHASE_COLUMNS]

    def get_phase_values(self, phase):
        return {key: fn(phase) for key, fn in self.TEST_PHASE_COLUMNS}

    def write_test_measurements(self, spreadsheet, test_record):
        """ Writes the 'Measurements' tab. """
        all_measures = []
        for phase in test_record.phases:
            for measure_name, measure in phase.measurements.items():
                measure_values = self.get_measurements_values(measure)
                measure_values['phase'] = phase.name
                all_measures.append(measure_values)

        tab = Worksheet(spreadsheet, MEASUREMENTS_TAB, headers=self.get_measurements_header())
        tab.append_rows(all_measures)

    def get_measurements_header(self):
        return ['phase', 'name', 'measured_value', 'outcome', 'validators']

    def get_measurements_values(self, measurement):
        measure = measurement.as_base_types()
        measure['validators'] = ' AND '.join(measure['validators'])
        return measure

    def write_test_config(self, spreadsheet, test_record):
        """ Writes the 'Config' tab. """
        header = ['name', 'value']
        
        tab = Worksheet(spreadsheet, CONFIG_TAB, headers=header)
        tab.append_rows([{'name': key, 'value': value} for key, value in test_record.metadata['config'].items()])

    def write_logs(self, spreadsheet, test_record):
        """ Writes the 'Logs' tab. """
        header = ['timestamp', 'timestamp_delta', 'logger_name', 'message']
        logs = []
        previous = None

        for lr in test_record.log_records:
            timestamp = millis_to_datetime(lr.timestamp_millis)
            if previous is None:
                delta = 0
            else:
                delta = (lr.timestamp_millis - previous)/1000
            
            previous = lr.timestamp_millis
            
            logs.append(
                {
                    'timestamp': str(timestamp),
                    'timestamp_delta': delta,
                    'logger_name': lr.logger_name,
                    'message': lr.message
                }
            )
        
        tab = Worksheet(spreadsheet, LOGS_TAB, headers=header)
        tab.append_rows(logs)
        
class TestRecordToRowOutput(object):
    """ Maintains an index file of all occured tests in a specific folder. """

    COLUMNS = [
        ('url', lambda tr: get_sheet_url(tr)),
        ('station_id', lambda tr: tr.station_id),
        ('dut_id', lambda tr: tr.dut_id),
        ('start_time', lambda tr: str(millis_to_datetime(tr.start_time_millis))),
        ('end_time', lambda tr: str(millis_to_datetime(tr.end_time_millis))),
        ('outcome', lambda tr: tr.outcome.name),
        ('error', lambda tr: format_error(tr))
    ]

    def __init__(self, folder, sheet_name):
        sheet = folder.get_or_create_spreadsheet(sheet_name)

        self.sheet = Worksheet(sheet, TESTS_TAB, headers=self.get_header(), first=True)
        self.measurements_sheet = Worksheet(sheet, MEASUREMENTS_TAB, headers=self.get_header())

    def __call__(self, test_record):
        self.add_test_row(test_record)
    
    def add_test_row(self, test_record):
        self.sheet.append_row(self.get_values(test_record))

    def get_header(self):
        return [col[0] for col in self.COLUMNS]

    def get_values(self, tr):
        return {key: fn(tr) for key, fn in self.COLUMNS}
        
def millis_to_datetime(millis):
    return datetime.datetime.utcfromtimestamp(millis/1000.0)

def next_available_row(worksheet):
    str_list = list(filter(lambda x: bool(x), worksheet.col_values(1)))
    return str(len(str_list)+1)

def format_error(test_record):
    errors = [':'.join(details) for details in test_record.outcome_details]
    return ' AND '.join(errors)

def get_sheet_url(test_record):
    file_url = test_record.metadata['google_sheet_result_file_url']
    return '=HYPERLINK("%s", "Link")' % file_url


"""Module to display test summary on console."""

import os
import sys

from openhtf.core import test_record
from openhtf.core import measurements
import six


class ConsoleSummary():
  """Print test results with failure info on console. """

  # pylint: disable=invalid-name
  def __init__(self, indent=2, output_stream=sys.stdout):
    self.indent = ' ' * indent
    if os.name == 'posix':    #Linux and Mac
      self.RED = '\033[91m'
      self.GREEN = '\033[92m'
      self.ORANGE = '\033[93m'
      self.RESET = '\033[0m'
      self.BOLD = '\033[1m'
    else:
      self.RED = ""
      self.GREEN = ""
      self.ORANGE = ""
      self.RESET = ""
      self.BOLD = ""

    self.color_table = {
        test_record.Outcome.PASS:self.GREEN,
        test_record.Outcome.FAIL:self.RED,
        test_record.Outcome.ERROR:self.ORANGE,
        test_record.Outcome.TIMEOUT:self.ORANGE,
        test_record.Outcome.ABORTED:self.RED,
    }
    self.output_stream = output_stream
  # pylint: enable=invalid-name

  def __call__(self, record):
    output_lines = [''.join((self.color_table[record.outcome],
                             self.BOLD, record.code_info.name, ':',
                             record.outcome.name, self.RESET))]
    if record.outcome != test_record.Outcome.PASS:
      for phase in record.phases:
        new_phase = True
        phase_time_sec = (float(phase.end_time_millis)
                          - float(phase.start_time_millis)) / 1000.0
        for name, measurement in six.iteritems(phase.measurements):
          if measurement.outcome != measurements.Outcome.PASS:
            if new_phase:
              output_lines.append('failed phase: %s [ran for %.2f sec]' %
                                  (phase.name, phase_time_sec))
              new_phase = False

            output_lines.append('%sfailed_item: %s (%s)' %
                                (self.indent, name, measurement.outcome))
            output_lines.append('%smeasured_value: %s' %
                                (self.indent*2, measurement.measured_value))
            output_lines.append('%svalidators:' % (self.indent*2))
            for validator in measurement.validators:
              output_lines.append('%svalidator: %s' %
                                  (self.indent*3, str(validator)))

        phase_result = phase.result.phase_result
        if not phase_result:   #Timeout
          output_lines.append('timeout phase: %s [ran for %.2f sec]' %
                              (phase.name, phase_time_sec))
        elif 'CONTINUE' not in str(phase_result):  #Exception
          output_lines.append('%sexception type: %s' %
                              (self.indent, record.outcome_details[0].code))

    output_lines.append('\n')
    text = '\n'.join(output_lines)
    self.output_stream.write(text)



"""Utilities for handling command line arguments.

StoreInModule:
  Enables emulating a gflags-esque API (flag affects global value), but one
  doesn't necessarily need to use flags to set values.

  Example usage:
    DEFAULT_VALUE = 0
    ARG_PARSER = argv.ModuleParser()
    ARG_PARSER.add_argument(
        '--override-value', action=argv.StoreInModule,
        default=DEFAULT_VALUE, target='%s.DEFAULT_VALUE' % __name__)

  Then in an entry point (main() function), use that parser as a parent:
    parser = argparse.ArgumentParser(parents=[other_module.ARG_PARSER])
    parser.parse_args()
"""

import argparse
import sys


def ModuleParser():
  return argparse.ArgumentParser(add_help=False)


class StoreInModule(argparse.Action):

  def __init__(self, *args, **kwargs):
    self._tgt_mod, self._tgt_attr = kwargs.pop('target').rsplit('.', 1)
    proxy_cls = kwargs.pop('proxy', None)
    if proxy_cls is not None:
      self._proxy = proxy_cls(*args, **kwargs)
    super(StoreInModule, self).__init__(*args, **kwargs)

  def __call__(self, parser, namespace, values, option_string=None):
    if hasattr(self, '_proxy'):
      values = self._proxy(parser, namespace, values)
    setattr(self._resolve_module(), self._tgt_attr, values)
    # self.val = values

  def _resolve_module(self):
    if '.' in self._tgt_mod:
      base, mod = self._tgt_mod.rsplit('.', 1)
      __import__(base, fromlist=[mod])
      return sys.modules[self._tgt_mod]
    else:
      return __import__(self._tgt_mod)


class _StoreValueInModule(StoreInModule):
  """Stores a value in a module level variable when set."""

  def __init__(self, const, *args, **kwargs):
    kwargs.update(nargs=0, const=const)
    super(_StoreValueInModule, self).__init__(*args, **kwargs)

  def __call__(self, parser, namespace, values, option_string=None):
    del values  # Unused.
    super(_StoreValueInModule, self).__call__(
        parser, namespace, self.const, option_string=option_string)


class StoreTrueInModule(_StoreValueInModule):
  """Stores True in a module level variable when set."""

  def __init__(self, *args, **kwargs):
    super(StoreTrueInModule, self).__init__(True, *args, **kwargs)


class StoreFalseInModule(_StoreValueInModule):
  """Stores False in module level variable when set."""

  def __init__(self, *args, **kwargs):
    super(StoreFalseInModule, self).__init__(False, *args, **kwargs)


class StoreRepsInModule(StoreInModule):
  """Store a count of number of times the flag was repeated in a module."""

  def __init__(self, *args, **kwargs):
    kwargs.update(nargs=0, const=None)
    super(StoreRepsInModule, self).__init__(*args, **kwargs)

  def __call__(self, parser, namespace, values, option_string=None):
    del values  # Unused.
    old_count = getattr(self._resolve_module(), self._tgt_attr)
    if old_count is None:
      super(StoreRepsInModule, self).__call__(
              parser, namespace, 0, option_string=option_string)
    else:
      super(StoreRepsInModule, self).__call__(
              parser, namespace, old_count + 1, option_string=option_string)



import paramiko
import socket
import time

import typing

from functools import wraps
from collections.abc import Sequence
from dataclasses import dataclass

from .base import UnboundPlug

class SSHError(Exception):
    pass

class SSHTimeoutError(SSHError):
    pass

def _ssh_client_connected(function):
    @wraps(function)
    def check_connected(*args, **kwargs):
        _self = args[0]
        if not _self.is_connected():
            _self.logger.info("Connection is dead, reopening...")
            _self.open()
        return function(*args, **kwargs)

    return check_connected

class SSHInterface(UnboundPlug):
    """ An interface to an SSH Server.
    """
    @dataclass()
    class SSHResponse():
        exit_code: int #: The command exit code
        err_output: str #: The command stderr output
        std_output: str #: The command stdout output
        
        @property
        def output(self):
            """Combines both the err output and the std_output."""
            return self.err_output + '\n' + self.std_output

    def __init__(self, addr, username, password, create_timeout=3, port=22):
        super().__init__()
        self.ssh = None
        self.addr = addr
        self.username = username
        self.password = password
        self.create_timeout = create_timeout
        self.port = port

    def open(self, _client=None):
        self.logger.info("(Initiating SSH connection at %s)", self.addr)
        self.logger.info("(addr={}:{}, user={!r}, password={!r})".format(self.addr, self.port, self.username, self.password))
        try:
            # _client allows to pass in a mock for testing
            if _client is None: _client = paramiko.SSHClient()
            self.ssh = _client
            self.ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
            self.ssh.connect(self.addr, port=self.port, username=self.username, password=self.password, timeout=self.create_timeout)
            self.ssh.get_transport().set_keepalive(5) #Send keepalive packet every 5 seconds
        except Exception as e:
            raise SSHError("%s %s" % (str(e), e.message))

    
    def is_connected(self):
        if self.ssh is None or self.ssh.get_transport() is None:
            return False
        else:
            return self.ssh.get_transport().is_active()

    def close(self):
        self.logger.info("(Closing SSH connection)")
        self.ssh.close()

    @_ssh_client_connected
    def execute_command(self, 
            command:str, 
            timeout:float=60, 
            stdin:typing.List=[], 
            get_pty:bool=False, 
            assertexitcode:typing.Union[typing.List[int], int, None]=0
        ):
        """ Send a :obj:`command` and wait for it to execute.

        Args:
            command: The command to send. End of lines are automatically managed. For example execute_command('ls')
                will executed the ls command.
            timeout: The timeout in second to wait for the command to finish executing.
            stdin: A list of inputs to send into stdin after the command is started. Each entry in this list will
                be separated with an r'\\n' character.
            get_pty: Usually required when providing :obj:`stdin`.
            assertexitcode: Unless this is None, defines one or a list of exit codes that are expected. After the 
                command is executed, an :class:`SSHError` will be raised if the exit code is not as expected.

        Raises:
            SSHTimeoutError:
                Raised when :obj:`timeout` is reached.
            SSHError:
                Raised when the exit code of the command is not in :obj:`assertexitcode` and :obj:`assertexitcode` is not None.
        """
        output = ""
        err_output = ""
        exit_code = None
        if command != "":
            self.logger.info("(Timeout %.1fs, TTY=%s)" % (timeout, str(get_pty)))
            ssh_stdin, ssh_stdout, ssh_stderr = self.ssh.exec_command(command, get_pty=get_pty)
    
            self.logger.debug("> {!r}".format(command))
            for stdin_element in stdin:
                self.logger.debug("> {!r}".format(stdin_element))
                ssh_stdin.write('%s\n'%stdin_element)
                ssh_stdin.flush()

            ssh_stdin.close()# Sends eof

            try:
                self.wait_stdout_with_timeout(ssh_stdout, timeout)
                exit_code = ssh_stdout.channel.recv_exit_status()
            finally:
                ssh_stdout.channel.close()
                ssh_stderr.channel.close()
                output = ssh_stdout.read().decode('utf-8', 'ignore')
                err_output = ssh_stderr.read().decode('utf-8', 'ignore')


            self.logger.debug("(Exit code={}, Response:)\n{}".format(exit_code, output.strip()))
            if err_output:
                self.logger.debug("(STDERR:)\n{}".format(err_output.strip()))
        else:
            pass
        
        if assertexitcode is not None:
            assert_exit_code(exit_code, expected=assertexitcode)

        response = self.SSHResponse(exit_code=exit_code, err_output=err_output, std_output=output)

        return response
    
    def wait_stdout_with_timeout(self, stdout, timeout_seconds):
        start_time = time.time()
        while time.time() - start_time < timeout_seconds:
            if stdout.channel.eof_received:
                break
            time.sleep(0)
        else:
            raise SSHTimeoutError(f'Client command timeout reached ({timeout_seconds}s)')



def assert_exit_code(exit_code, expected):
    if not isinstance(expected, Sequence):
        expected = [expected]
        
    if exit_code not in expected:
        raise SSHError('Exit code {} not in expected list {}'.format(exit_code, expected))
        




# Copyright 2014 Google Inc. All Rights Reserved.

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""The main OpenHTF entry point."""

import pkg_resources
import signal

from openhtf import plugs
from openhtf.core import phase_executor
from openhtf.core import test_record
from openhtf.core.measurements import Dimension
from openhtf.core.measurements import Measurement
from openhtf.core.measurements import measures
from openhtf.core.monitors import monitors
from openhtf.core.phase_descriptor import PhaseDescriptor
from openhtf.core.phase_descriptor import PhaseOptions
from openhtf.core.phase_descriptor import PhaseResult
from openhtf.core.phase_group import PhaseGroup
from openhtf.core.test_descriptor import Test
from openhtf.core.test_descriptor import TestApi
from openhtf.core.test_descriptor import TestDescriptor
from openhtf.plugs import plug
from openhtf.util import conf
from openhtf.util import console_output
from openhtf.util import data
from openhtf.util import functions
from openhtf.util import logs
from openhtf.util import units

# TODO: TestPhase is used for legacy reasons and should be deprecated.
TestPhase = PhaseOptions  # pylint: disable=invalid-name


def get_version():
  """Returns the version string of the 'openhtf' package.

  Note: the version number doesn't seem to get properly set when using ipython.
  """
  try:
    return pkg_resources.get_distribution('openhtf')
  except pkg_resources.DistributionNotFound:
    return 'Unknown - Perhaps openhtf was not installed via setup.py or pip.'

__version__ = get_version()

# Register signal handler to stop all tests on SIGINT.
signal.signal(signal.SIGINT, Test.handle_sig_int)
